from django.urls import path

from projectroles import views, views_ajax, views_api

app_name = 'projectroles'

# UI views
urls_ui = [
    path(
        route='<uuid:project>',
        view=views.ProjectDetailView.as_view(),
        name='detail',
    ),
    path(
        route='update/<uuid:project>',
        view=views.ProjectUpdateView.as_view(),
        name='update',
    ),
    path(
        route='create',
        view=views.ProjectCreateView.as_view(),
        name='create',
    ),
    path(
        route='create/<uuid:project>',
        view=views.ProjectCreateView.as_view(),
        name='create',
    ),
    path(
        route='archive/<uuid:project>',
        view=views.ProjectArchiveView.as_view(),
        name='archive',
    ),
    path(
        route='delete/<uuid:project>',
        view=views.ProjectDeleteView.as_view(),
        name='delete',
    ),
    # Search views
    path(
        route='search/results/',
        view=views.ProjectSearchResultsView.as_view(),
        name='search',
    ),
    path(
        route='search/advanced',
        view=views.ProjectAdvancedSearchView.as_view(),
        name='search_advanced',
    ),
    # Project role views
    path(
        route='members/<uuid:project>',
        view=views.ProjectRoleView.as_view(),
        name='roles',
    ),
    path(
        route='members/<uuid:project>',
        view=views.ProjectRoleView.as_view(),
        name='roles',
    ),
    path(
        route='members/create/<uuid:project>',
        view=views.RoleAssignmentCreateView.as_view(),
        name='role_create',
    ),
    path(
        route='members/promote/<uuid:project>/assignment/<uuid:promote_as>',
        view=views.RoleAssignmentCreateView.as_view(),
        name='role_create_promote',
    ),
    path(
        route='members/update/<uuid:roleassignment>',
        view=views.RoleAssignmentUpdateView.as_view(),
        name='role_update',
    ),
    path(
        route='members/delete/<uuid:roleassignment>',
        view=views.RoleAssignmentDeleteView.as_view(),
        name='role_delete',
    ),
    path(
        route='members/delete/own/<uuid:roleassignment>',
        view=views.RoleAssignmentOwnDeleteView.as_view(),
        name='role_delete_own',
    ),
    path(
        route='members/owner/transfer/<uuid:project>',
        view=views.RoleAssignmentOwnerTransferView.as_view(),
        name='role_owner_transfer',
    ),
    # Project invite views
    path(
        route='invites/<uuid:project>',
        view=views.ProjectInviteView.as_view(),
        name='invites',
    ),
    path(
        route='invites/create/<uuid:project>',
        view=views.ProjectInviteCreateView.as_view(),
        name='invite_create',
    ),
    # Workaround for issue #1510
    path(
        route='invites/create/<uuid:project>?e=<str:email>&r=<int:role>',
        view=views.ProjectInviteCreateView.as_view(),
        name='invite_create',
    ),
    path(
        route='invites/accept/<str:secret>',
        view=views.ProjectInviteAcceptView.as_view(),
        name='invite_accept',
    ),
    path(
        route='invites/process/login/<str:secret>',
        view=views.ProjectInviteProcessLoginView.as_view(),
        name='invite_process_login',
    ),
    path(
        route='invites/process/new-user/<str:secret>',
        view=views.ProjectInviteProcessNewUserView.as_view(),
        name='invite_process_new_user',
    ),
    path(
        route='invites/resend/<uuid:projectinvite>',
        view=views.ProjectInviteResendView.as_view(),
        name='invite_resend',
    ),
    path(
        route='invites/revoke/<uuid:projectinvite>',
        view=views.ProjectInviteRevokeView.as_view(),
        name='invite_revoke',
    ),
    path(
        route='user/update',
        view=views.UserUpdateView.as_view(),
        name='user_update',
    ),
    path(
        route='site-app-settings',
        view=views.SiteAppSettingsView.as_view(),
        name='site_app_settings',
    ),
    # Remote site and project views
    path(
        route='remote/sites',
        view=views.RemoteSiteListView.as_view(),
        name='remote_sites',
    ),
    path(
        route='remote/site/add',
        view=views.RemoteSiteCreateView.as_view(),
        name='remote_site_create',
    ),
    path(
        route='remote/site/update/<uuid:remotesite>',
        view=views.RemoteSiteUpdateView.as_view(),
        name='remote_site_update',
    ),
    path(
        route='remote/site/delete/<uuid:remotesite>',
        view=views.RemoteSiteDeleteView.as_view(),
        name='remote_site_delete',
    ),
    path(
        route='remote/site/<uuid:remotesite>',
        view=views.RemoteProjectListView.as_view(),
        name='remote_projects',
    ),
    path(
        route='remote/site/access/<uuid:remotesite>',
        view=views.RemoteProjectBatchUpdateView.as_view(),
        name='remote_projects_update',
    ),
    path(
        route='remote/site/sync/<uuid:remotesite>',
        view=views.RemoteProjectSyncView.as_view(),
        name='remote_projects_sync',
    ),
]

# Ajax API views
urls_ajax = [
    path(
        route='ajax/list',
        view=views_ajax.ProjectListAjaxView.as_view(),
        name='ajax_project_list',
    ),
    path(
        route='ajax/list/columns',
        view=views_ajax.ProjectListColumnAjaxView.as_view(),
        name='ajax_project_list_columns',
    ),
    path(
        route='ajax/list/roles',
        view=views_ajax.ProjectListRoleAjaxView.as_view(),
        name='ajax_project_list_roles',
    ),
    path(
        route='ajax/stats/category/<uuid:project>',
        view=views_ajax.CategoryStatisticsAjaxView.as_view(),
        name='ajax_stats_category',
    ),
    path(
        route='ajax/star/<uuid:project>',
        view=views_ajax.ProjectStarringAjaxView.as_view(),
        name='ajax_star',
    ),
    path(
        route='ajax/star/home',
        view=views_ajax.HomeStarringAjaxView.as_view(),
        name='ajax_star_home',
    ),
    path(
        route='ajax/remote/access/<uuid:project>',
        view=views_ajax.RemoteProjectAccessAjaxView.as_view(),
        name='ajax_remote_access',
    ),
    path(
        route='ajax/sidebar/<uuid:project>',
        view=views_ajax.SidebarContentAjaxView.as_view(),
        name='ajax_sidebar',
    ),
    path(
        route='ajax/dropdown',
        view=views_ajax.UserDropdownContentAjaxView.as_view(),
        name='ajax_user_dropdown',
    ),
    path(
        route='ajax/settings/site-read-only',
        view=views_ajax.SiteReadOnlySettingAjaxView.as_view(),
        name='ajax_settings_site_read_only',
    ),
    path(
        route='ajax/user/current',
        view=views_ajax.CurrentUserRetrieveAjaxView.as_view(),
        name='ajax_user_current',
    ),
    path(
        route='ajax/autocomplete/user',
        view=views_ajax.UserAutocompleteAjaxView.as_view(),
        name='ajax_autocomplete_user',
    ),
    path(
        route='ajax/autocomplete/user/redirect',
        view=views_ajax.UserAutocompleteRedirectAjaxView.as_view(
            create_field='user'
        ),
        name='ajax_autocomplete_user_redirect',
    ),
]

# REST API views
urls_api = [
    path(
        route='api/list',
        view=views_api.ProjectListAPIView.as_view(),
        name='api_project_list',
    ),
    path(
        route='api/retrieve/<uuid:project>',
        view=views_api.ProjectRetrieveAPIView.as_view(),
        name='api_project_retrieve',
    ),
    path(
        route='api/create',
        view=views_api.ProjectCreateAPIView.as_view(),
        name='api_project_create',
    ),
    path(
        route='api/update/<uuid:project>',
        view=views_api.ProjectUpdateAPIView.as_view(),
        name='api_project_update',
    ),
    path(
        route='api/destroy/<uuid:project>',
        view=views_api.ProjectDestroyAPIView.as_view(),
        name='api_project_destroy',
    ),
    path(
        route='api/roles/create/<uuid:project>',
        view=views_api.RoleAssignmentCreateAPIView.as_view(),
        name='api_role_create',
    ),
    path(
        route='api/roles/update/<uuid:roleassignment>',
        view=views_api.RoleAssignmentUpdateAPIView.as_view(),
        name='api_role_update',
    ),
    path(
        route='api/roles/destroy/<uuid:roleassignment>',
        view=views_api.RoleAssignmentDestroyAPIView.as_view(),
        name='api_role_destroy',
    ),
    path(
        route='api/roles/owner-transfer/<uuid:project>',
        view=views_api.RoleAssignmentOwnerTransferAPIView.as_view(),
        name='api_role_owner_transfer',
    ),
    path(
        route='api/invites/list/<uuid:project>',
        view=views_api.ProjectInviteListAPIView.as_view(),
        name='api_invite_list',
    ),
    path(
        route='api/invites/retrieve/<uuid:projectinvite>',
        view=views_api.ProjectInviteRetrieveAPIView.as_view(),
        name='api_invite_retrieve',
    ),
    path(
        route='api/invites/create/<uuid:project>',
        view=views_api.ProjectInviteCreateAPIView.as_view(),
        name='api_invite_create',
    ),
    path(
        route='api/invites/revoke/<uuid:projectinvite>',
        view=views_api.ProjectInviteRevokeAPIView.as_view(),
        name='api_invite_revoke',
    ),
    path(
        route='api/invites/resend/<uuid:projectinvite>',
        view=views_api.ProjectInviteResendAPIView.as_view(),
        name='api_invite_resend',
    ),
    path(
        route='api/settings/retrieve/<uuid:project>',
        view=views_api.ProjectSettingRetrieveAPIView.as_view(),
        name='api_project_setting_retrieve',
    ),
    path(
        route='api/settings/set/<uuid:project>',
        view=views_api.ProjectSettingSetAPIView.as_view(),
        name='api_project_setting_set',
    ),
    path(
        route='api/settings/retrieve/user',
        view=views_api.UserSettingRetrieveAPIView.as_view(),
        name='api_user_setting_retrieve',
    ),
    path(
        route='api/settings/set/user',
        view=views_api.UserSettingSetAPIView.as_view(),
        name='api_user_setting_set',
    ),
    path(
        route='api/users/list',
        view=views_api.UserListAPIView.as_view(),
        name='api_user_list',
    ),
    path(
        route='api/users/<uuid:user>',
        view=views_api.UserRetrieveAPIView.as_view(),
        name='api_user_retrieve',
    ),
    path(
        route='api/users/current',
        view=views_api.CurrentUserRetrieveAPIView.as_view(),
        name='api_user_current',
    ),
    path(
        route='api/remote/get/<str:secret>',
        view=views_api.RemoteProjectGetAPIView.as_view(),
        name='api_remote_get',
    ),
]

urlpatterns = urls_ui + urls_ajax + urls_api



"""
Blockprojectaccess management command for temporarily blocking or unblocking
user access to a project.
"""

import sys

from django.core.management.base import BaseCommand

from projectroles.app_settings import AppSettingAPI
from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import Project


app_settings = AppSettingAPI()
logger = ManagementCommandLogger(__name__)


# Local constants
APP_NAME = 'projectroles'
SET_NAME = 'project_access_block'
PROJECT_NOT_FOUND_MSG = 'Project not found with UUID: {project_uuid}'
INVALID_PROJECT_TYPE_MSG = 'This action only supports projects of PROJECT type'
MODE_CONVERT_ERR_MSG = 'Exception converting mode "{mode}": {ex}'
INVALID_MODE_MSG = 'Invalid mode: {mode} (accepted: 0, 1)'


class Command(BaseCommand):
    """Command for cleaning up undefined or orphaned app settings"""

    help = (
        'Toggle temporary project access block. When enabled, this prevents '
        'access to all views to the given project for non-superusers.'
    )

    def add_arguments(self, parser):
        parser.add_argument(
            '-p',
            '--project',
            dest='project',
            required=True,
            action='store',
            help='UUID of project to be blocked or unblocked',
        )
        parser.add_argument(
            '-m',
            '--mode',
            dest='mode',
            required=True,
            action='store',
            help='Mode to be set (1=blocked, 0=unblocked)',
        )

    def handle(self, *args, **options):
        p_uuid = options.get('project')
        project = Project.objects.filter(sodar_uuid=p_uuid).first()
        if not project:
            logger.error(PROJECT_NOT_FOUND_MSG.format(project_uuid=p_uuid))
            sys.exit(1)
        if project.is_category():
            logger.error(INVALID_PROJECT_TYPE_MSG)
            sys.exit(1)
        try:
            mode = int(options.get('mode'))
        except Exception as ex:
            logger.error(
                MODE_CONVERT_ERR_MSG.format(mode=options.get('mode'), ex=ex)
            )
            sys.exit(1)
        if mode not in [0, 1]:
            logger.error(INVALID_MODE_MSG.format(mode=mode))
            sys.exit(1)
        old_val = app_settings.get(APP_NAME, SET_NAME, project=project)
        logger.debug(f'Previous mode: {int(old_val)}')
        app_settings.set(APP_NAME, SET_NAME, value=bool(mode), project=project)
        action = 'enabled' if mode else 'disabled'
        logger.info(f'Project access block {action}: {project.get_log_title()}')



"""Ajax views for the adminalerts app"""

from django.http import HttpResponseBadRequest

from rest_framework.response import Response

# Projectroles dependency
from projectroles.views_ajax import SODARBasePermissionAjaxView

from adminalerts.models import AdminAlert


class AdminAlertActiveToggleAjaxView(SODARBasePermissionAjaxView):
    """AdminAlert acivation toggling Ajax view"""

    permission_required = 'adminalerts.update_alert'
    http_method_names = ['post']

    def post(self, request, **kwargs):
        alert_uuid = kwargs.get('adminalert', None)
        try:
            alert = AdminAlert.objects.get(sodar_uuid__exact=alert_uuid)
        except AdminAlert.DoesNotExist:
            return HttpResponseBadRequest()
        alert.active = not alert.active
        alert.save()
        return Response({'is_active': alert.active}, status=200)



"""Tests for UI view permissions in the filesfolders app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_permissions import ProjectPermissionTestBase

from filesfolders.models import File, Folder
from filesfolders.tests.test_models import (
    FileMixin,
    FolderMixin,
    HyperLinkMixin,
)


app_settings = AppSettingAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME = 'filesfolders'
SECRET = '7dqq83clo2iyhg29hifbor56og6911r5'


# Base Classes and Mixins ------------------------------------------------------


class FilesfoldersPermissionTestMixin(FolderMixin, FileMixin, HyperLinkMixin):
    """Mixin for filesfolders view permission test helpers"""

    def make_test_folder(self) -> Folder:
        return self.make_folder(
            name='folder',
            project=self.project,
            folder=None,
            owner=self.user_owner,  # Project owner is the owner of folder
            description='',
        )

    def make_test_file(self, public: bool = False) -> File:
        if public:
            app_settings.set(
                APP_NAME, 'allow_public_links', True, project=self.project
            )
        return self.make_file(
            name='file.txt',
            file_name='file.txt',
            file_content=bytes('content'.encode('utf-8')),
            project=self.project,
            folder=None,
            owner=self.user_owner,  # Project owner is the file owner
            description='',
            public_url=public,
            secret=SECRET,
        )

    def make_test_link(self):
        return self.make_hyperlink(
            name='Link',
            url='https://www.google.com/',
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='',
        )


# Test Cases -------------------------------------------------------------------


class TestProjectFileView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for ProjectFileView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:list', kwargs={'project': self.project.sodar_uuid}
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test ProjectFileView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        # Test public project
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)


class TestFolderCreateView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FolderCreateView view permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:folder_create',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_get(self):
        """Test FolderCreateView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_category(self):
        """Test GET with folder under category (should fail)"""
        url = reverse(
            'filesfolders:folder_create',
            kwargs={'project': self.category.sodar_uuid},
        )
        self.assert_response(url, self.all_users, 302)


class TestFolderUpdateView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FolderUpdateView permissions"""

    def setUp(self):
        super().setUp()
        folder = self.make_test_folder()
        self.url = reverse(
            'filesfolders:folder_update',
            kwargs={'item': folder.sodar_uuid},
        )

    def test_get(self):
        """Test FolderUpdateView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,  # NOTE: not the owner of the folder
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)


class TestFolderDeleteView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FolderDeleteView permissions"""

    def setUp(self):
        super().setUp()
        folder = self.make_test_folder()
        self.url = reverse(
            'filesfolders:folder_delete',
            kwargs={'item': folder.sodar_uuid},
        )

    def test_get(self):
        """Test FolderDeleteView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)


class TestFileCreateView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FileCreateView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:file_create',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_get(self):
        """Test FileCreateView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_category(self):
        """Test GET under category (should fail)"""
        url = reverse(
            'filesfolders:file_create',
            kwargs={'project': self.category.sodar_uuid},
        )
        self.assert_response(url, self.all_users, 302)


class TestFileUpdateView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FileUpdateView permissions"""

    def setUp(self):
        super().setUp()
        file = self.make_test_file()
        self.url = reverse(
            'filesfolders:file_update', kwargs={'item': file.sodar_uuid}
        )

    def test_get(self):
        """Test FileUpdateView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_owner,
            self.user_delegate_cat,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,  # NOTE: not the owner of the file
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)


class TestFileDeleteView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FileDeleteView permissions"""

    def setUp(self):
        super().setUp()
        file = self.make_test_file()
        self.url = reverse(
            'filesfolders:file_delete', kwargs={'item': file.sodar_uuid}
        )

    def test_get(self):
        """Test FileDeleteView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_owner,
            self.user_delegate_cat,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)


class TestFilePublicLinkView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FilePublicLinkView permissions"""

    def setUp(self):
        super().setUp()
        file = self.make_test_file(public=True)
        self.url = reverse(
            'filesfolders:file_public_link',
            kwargs={'file': file.sodar_uuid},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        self.bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test FilePublicLinkView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)


class TestFileServeView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FileServeView permissions"""

    def setUp(self):
        super().setUp()
        file = self.make_test_file()
        self.url = reverse(
            'filesfolders:file_serve',
            kwargs={'file': file.sodar_uuid, 'file_name': file.name},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_finder_cat,
            self.user_viewer_cat,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test FileServeView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)


class TestFileServePublicView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for FileServePublicView permissions"""

    def setUp(self):
        super().setUp()
        file = self.make_test_file(public=True)
        self.url = reverse(
            'filesfolders:file_serve_public',
            kwargs={'secret': SECRET, 'file_name': file.name},
        )

    def test_get(self):
        """Test FileServePublicView GET"""
        self.assert_response(self.url, self.all_users, 200)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.all_users, 200)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.all_users, 200)

    def test_get_disabled(self):
        """Test GET with public links disabled (should fail)"""
        app_settings.set(
            APP_NAME, 'allow_public_links', False, project=self.project
        )
        self.assert_response(self.url, self.all_users, 400)


class TestHyperLinkCreateView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for HyperLinkCreateView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'filesfolders:hyperlink_create',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_get(self):
        """Test HyperLinkCreateView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_category(self):
        """Test GET under category (should fail)"""
        url = reverse(
            'filesfolders:hyperlink_create',
            kwargs={'project': self.category.sodar_uuid},
        )
        self.assert_response(url, self.all_users, 302)


class TestHyperLinkUpdateView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for HyperLinkUpdateView permissions"""

    def setUp(self):
        super().setUp()
        link = self.make_test_link()
        self.url = reverse(
            'filesfolders:hyperlink_update',
            kwargs={'item': link.sodar_uuid},
        )

    def test_get(self):
        """Test HyperLinkUpdateView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_owner,
            self.user_delegate_cat,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test hyperlink updating with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)


class TestHyperLinkDeleteView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for HyperLinkDeleteView permissions"""

    def setUp(self):
        super().setUp()
        link = self.make_test_link()
        self.url = reverse(
            'filesfolders:hyperlink_update',
            kwargs={'item': link.sodar_uuid},
        )

    def test_get(self):
        """Test HyperLinkDeleteView GET"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_owner,
            self.user_delegate_cat,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,  # NOTE: not the owner of the link
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)


class TestBatchEditView(
    FilesfoldersPermissionTestMixin, ProjectPermissionTestBase
):
    """Tests for BatchEditView permissions"""

    def setUp(self):
        super().setUp()
        folder = self.make_test_folder()
        self.url = reverse(
            'filesfolders:batch_edit',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.post_data = {
            'batch-action': 'delete',
            'user-confirmed': '0',
            f'batch_item_Folder_{folder.sodar_uuid}': '1',
        }

    def test_post(self):
        """Test BatchEditView POST"""
        # NOTE: Contributor is OK as checks for object perms happen after POST
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.assert_response(
            self.url, good_users, 200, method='POST', data=self.post_data
        )
        self.assert_response(
            self.url, bad_users, 302, method='POST', data=self.post_data
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response(
            self.url, bad_users, 302, method='POST', data=self.post_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_anon(self):
        """Test POST with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 302)

    def test_post_archive(self):
        """Test POST with archived project"""
        self.project.set_archive()
        self.assert_response(
            self.url, self.superuser, 200, method='POST', data=self.post_data
        )
        self.assert_response(
            self.url,
            self.non_superusers,
            302,
            method='POST',
            data=self.post_data,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response(
            self.url,
            self.no_role_users,
            302,
            method='POST',
            data=self.post_data,
        )

    def test_post_block(self):
        """Test POST with project access block"""
        self.set_access_block(self.project)
        self.assert_response(
            self.url, self.superuser, 200, method='POST', data=self.post_data
        )
        self.assert_response(
            self.url,
            self.non_superusers,
            302,
            method='POST',
            data=self.post_data,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response(
            self.url,
            self.no_role_users,
            302,
            method='POST',
            data=self.post_data,
        )

    def test_post_read_only(self):
        """Test POST with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(
            self.url, self.superuser, 200, method='POST', data=self.post_data
        )
        self.assert_response(
            self.url,
            self.non_superusers,
            302,
            method='POST',
            data=self.post_data,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response(
            self.url,
            self.no_role_users,
            302,
            method='POST',
            data=self.post_data,
        )



"""Template tags for the filesfolders app"""

import logging

from typing import Any

from django import template
from django.conf import settings

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.models import Project

from filesfolders.models import File, HyperLink, FILESFOLDERS_FLAGS


app_settings = AppSettingAPI()
logger = logging.getLogger(__name__)
register = template.Library()


APP_NAME = 'filesfolders'


@register.filter
def get_class(obj: Any) -> str:
    return obj.__class__.__name__


@register.simple_tag
def get_details_items(project: Project) -> list:
    """Return recent files/links for card on project details page"""
    files = File.objects.filter(project=project).order_by('-date_modified')[:5]
    links = HyperLink.objects.filter(project=project).order_by(
        '-date_modified'
    )[:5]
    ret = list(files) + list(links)
    ret.sort(key=lambda x: x.date_modified, reverse=True)
    return ret[:5]


@register.simple_tag
def get_file_icon(file: File) -> str:
    """Return file icon"""
    ret = 'file-outline'
    try:
        mt = file.file.file.mimetype
    except Exception as ex:
        if settings.DEBUG:
            raise ex
        logger.error(
            f'Exception in accessing file data (UUID={file.sodar_uuid}): {ex}'
        )
        return ret
    if mt == 'application/pdf':
        ret = 'file-pdf-outline'
    elif (
        mt == 'application/vnd.openxmlformats-officedocument.'
        'presentationml.presentation'
    ):
        ret = 'file-powerpoint-outline'
    elif 'compressed' in mt or 'zip' in mt:
        ret = 'archive-outline'
    elif (
        'excel' in mt
        or mt == 'application/vnd.openxmlformats-'
        'officedocument.spreadsheetml.sheet'
    ):
        ret = 'file-excel-outline'
    elif 'image/' in mt:
        ret = 'file-image-outline'
    elif 'text/' in mt:
        ret = 'file-document-outline'
    # Default if not found
    return 'mdi:' + ret


@register.simple_tag
def get_flag(flag_name: str, tooltip: bool = True) -> str:
    """Return item flag HTML"""
    f = FILESFOLDERS_FLAGS[flag_name]
    tip_str = ''
    if tooltip:
        tip_str = (
            'title="{}" data-toggle="tooltip" '
            'data-placement="top"'.format(f['label'])
        )
    return (
        '<i class="iconify text-{} sodar-ff-flag-icon" data-icon="{}" {}>'
        '</i>'.format(f['color'], f['icon'], tip_str)
    )


@register.simple_tag
def get_flag_status(val: str, choice: str) -> str:
    """Return item flag status HTML for form"""
    if val == choice:
        return 'checked="checked"'
    return ''


@register.simple_tag
def get_flag_classes(flag_name: str) -> str:
    """Return CSS classes for item link based on flag name"""
    return FILESFOLDERS_FLAGS[flag_name]['text_classes']



"""View tests for the timeline app"""

from django.urls import reverse

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAPI

from timeline.tests.test_models import (
    TimelineEventTestBase,
    TimelineEventMixin,
    TimelineEventStatusMixin,
    EXTRA_DATA,
)


plugin_api = PluginAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME_PR = 'projectroles'


class ViewTestBase(
    TimelineEventMixin, TimelineEventStatusMixin, TimelineEventTestBase
):
    """Base class for timeline view testing"""

    def setUp(self):
        super().setUp()
        self.timeline = plugin_api.get_backend_api('timeline_backend')

        # Init superuser
        self.user = self.make_user('superuser')
        self.user.is_staff = True
        self.user.is_superuser = True
        self.user.save()
        # Init category and project
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self.make_assignment(
            self.category, self.user, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )

        # Init default event
        self.event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )


class TestProjectTimelineView(ViewTestBase):
    """Tests for ProjectTimelineView"""

    def test_get(self):
        """Test ProjectTimelineView GET"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:list_project',
                    kwargs={'project': self.project.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['object_list']), 1)
        self.assertEqual(response.context['object_list'].first(), self.event)

    def test_get_category(self):
        """Test GET with category"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:list_project',
                    kwargs={'project': self.category.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)


class TestProjectObjectTimelineView(ViewTestBase):
    """Tests for ProjectObjectTimelineView"""

    def setUp(self):
        super().setUp()
        # Add user as an object reference
        self.obj_ref = self.event.add_object(
            obj=self.user, label='user', name=self.user.username
        )

    def test_get(self):
        """Test ProjectObjectTimelineView GET"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:list_object',
                    kwargs={
                        'project': self.project.sodar_uuid,
                        'object_model': self.obj_ref.object_model,
                        'object_uuid': self.obj_ref.object_uuid,
                    },
                )
            )
        self.assertEqual(response.status_code, 200)


class TestSiteTimelineView(ViewTestBase):
    """Tests for SiteTimelineView"""

    def setUp(self):
        super().setUp()
        self.event_site = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.user,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )

    def test_get(self):
        """Test SiteTimelineView GET"""
        with self.login(self.user):
            response = self.client.get(reverse('timeline:list_site'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['object_list']), 1)
        self.assertEqual(
            response.context['object_list'].first(), self.event_site
        )


class TestSiteObjectTimelineView(ViewTestBase):
    """Tests for SiteObjectTimelineView"""

    def setUp(self):
        super().setUp()
        self.event_site = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.user,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )
        self.obj_ref = self.event_site.add_object(
            obj=self.user, label='user', name=self.user.username
        )

    def test_get(self):
        """Test SiteObjectTimelineView GET"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:list_object_site',
                    kwargs={
                        'object_model': self.obj_ref.object_model,
                        'object_uuid': self.obj_ref.object_uuid,
                    },
                )
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['object_list']), 1)
        self.assertEqual(
            response.context['object_list'].first(), self.event_site
        )


class TestAdminTimelineView(ViewTestBase):
    """Tests for AdminTimelineView"""

    def setUp(self):
        super().setUp()
        self.event_site = self.timeline.add_event(
            project=None,
            app_name=APP_NAME_PR,
            user=self.user,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )

    def test_get(self):
        """Test AdminTimelineView GET"""
        with self.login(self.user):
            response = self.client.get(reverse('timeline:list_admin'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['object_list']), 2)
        self.assertEqual(
            response.context['object_list'].first(), self.event_site
        )
        self.assertEqual(response.context['object_list'][1], self.event)



"""Tests for views in the siteinfo app"""

from django.urls import reverse

from test_plus.test import TestCase


class TestSiteInfoView(TestCase):
    """Tests for SiteInfoView"""

    def setUp(self):
        # Create users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user')
        # No user
        self.anonymous = None

    def test_get(self):
        """Test SiteInfoView GET"""
        with self.login(self.superuser):
            response = self.client.get(reverse('siteinfo:info'))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context['project_plugins'])
        self.assertIsNotNone(response.context['site_plugins'])
        self.assertIsNotNone(response.context['backend_plugins'])
        self.assertIsNotNone(response.context['settings_core'])



"""Models for the adminalerts app"""

import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone

from markupfield.fields import MarkupField


# Access Django user model
AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


class AdminAlert(models.Model):
    """
    An un-dismissable alert from a superuser to site users. Not dependent on
    project. Will expire after a set time.
    """

    #: Alert message to be shown for users
    message = models.CharField(
        max_length=255,
        unique=False,
        help_text='Alert message to be shown for users',
    )

    #: Superuser who has set the alert
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        related_name='alerts',
        help_text='Superuser who has set the alert',
        on_delete=models.CASCADE,
    )

    #: Full description (optional, will be shown on a separate page)
    description = MarkupField(
        unique=False,
        blank=True,
        null=True,
        markup_type='markdown',
        help_text='Full description of alert '
        '(optional, will be shown on a separate page)',
    )

    #: Alert creation timestamp
    date_created = models.DateTimeField(
        auto_now_add=True, help_text='Alert creation timestamp'
    )

    #: Alert expiration timestamp
    date_expire = models.DateTimeField(
        blank=False, null=False, help_text='Alert expiration timestamp'
    )

    #: Alert status (for disabling the alert before expiration)
    active = models.BooleanField(
        default=True,
        help_text='Alert status (for disabling the alert before expiration)',
    )

    #: Require authorization to view alert
    require_auth = models.BooleanField(
        default=True, help_text='Require authorization to view alert'
    )

    #: Adminalerts SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Adminalerts SODAR UUID'
    )

    def __str__(self):
        return '{}{}'.format(
            self.message,
            (
                ' [ACTIVE]'
                if (self.active and self.date_expire > timezone.now())
                else ''
            ),
        )

    def __repr__(self):
        values = (self.message, self.user.username, self.active)
        return 'AdminAlert({})'.format(', '.join(repr(v) for v in values))

    def is_active(self) -> bool:
        """Return True if alert is active and has not expired"""
        return (
            True
            if (self.date_expire > timezone.now() and self.active)
            else False
        )

    def is_expired(self) -> bool:
        """Return True if alert has expired"""
        return True if self.date_expire < timezone.now() else False



"""Tests for UI view permissions in the userprofile app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_models import (
    SODARUserAdditionalEmailMixin,
    ADD_EMAIL,
)
from projectroles.tests.test_permissions import SiteAppPermissionTestBase


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']


class TestUserProfilePermissions(
    SODARUserAdditionalEmailMixin, SiteAppPermissionTestBase
):
    """Tests for userprofile view permissions"""

    def setUp(self):
        super().setUp()
        self.regular_user2 = self.make_user('regular_user2')

    def test_get_profile(self):
        """Test UserDetailView GET"""
        url = reverse('userprofile:detail')
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_profile_anon(self):
        """Test UserDetailView GET with anonymous access"""
        url = reverse('userprofile:detail')
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, self.anonymous, 302)

    def test_get_profile_read_only(self):
        """Test UserDetailView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('userprofile:detail')
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, self.anonymous, 302)

    def test_get_settings_update(self):
        """Test UserSettingUpdateView GET"""
        url = reverse('userprofile:settings_update')
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_settings_update_anon(self):
        """Test UserSettingUpdateView GET with anonymous access"""
        url = reverse('userprofile:settings_update')
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, self.anonymous, 302)

    def test_get_settings_update_read_only(self):
        """Test UserSettingUpdateView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('userprofile:settings_update')
        self.assert_response(url, self.superuser, 200)
        self.assert_response(url, [self.regular_user, self.anonymous], 302)

    def test_get_email_create(self):
        """Test UserEmailCreateView GET"""
        url = reverse('userprofile:email_create')
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_email_create_anon(self):
        """Test UserEmailCreateView GET with anonymous access"""
        url = reverse('userprofile:email_create')
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, self.anonymous, 302)

    def test_get_email_create_read_only(self):
        """Test UserEmailCreateView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('userprofile:email_create')
        self.assert_response(url, self.superuser, 200)
        self.assert_response(url, [self.regular_user, self.anonymous], 302)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_get_email_create_target(self):
        """Test UserEmailCreateView GET as target site"""
        url = reverse('userprofile:email_create')
        self.assert_response(url, self.superuser, 200)
        self.assert_response(url, [self.regular_user, self.anonymous], 302)

    def test_get_email_delete(self):
        """Test UserEmailDeleteView GET"""
        email = self.make_email(self.regular_user, ADD_EMAIL)
        url = reverse(
            'userprofile:email_delete',
            kwargs={'sodaruseradditionalemail': email.sodar_uuid},
        )
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, [self.regular_user2, self.anonymous], 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_email_delete_anon(self):
        """Test UserEmailDeleteView GET with anonymous access"""
        email = self.make_email(self.regular_user, ADD_EMAIL)
        url = reverse(
            'userprofile:email_delete',
            kwargs={'sodaruseradditionalemail': email.sodar_uuid},
        )
        self.assert_response(url, [self.superuser, self.regular_user], 200)
        self.assert_response(url, [self.regular_user2, self.anonymous], 302)

    def test_get_email_delete_read_only(self):
        """Test UserEmailDeleteView GET with site read-only mode"""
        self.set_site_read_only()
        email = self.make_email(self.regular_user, ADD_EMAIL)
        url = reverse(
            'userprofile:email_delete',
            kwargs={'sodaruseradditionalemail': email.sodar_uuid},
        )
        bad_users = [self.regular_user, self.regular_user2, self.anonymous]
        self.assert_response(url, self.superuser, 200)
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_get_email_delete_target(self):
        """Test UserEmailDeleteView GET as target site"""
        email = self.make_email(self.regular_user, ADD_EMAIL)
        url = reverse(
            'userprofile:email_delete',
            kwargs={'sodaruseradditionalemail': email.sodar_uuid},
        )
        self.assert_response(url, [self.superuser], 200)
        self.assert_response(
            url, [self.regular_user, self.regular_user2, self.anonymous], 302
        )



# Generated by Django 4.2.14 on 2024-07-16 09:30

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


def populate_event_status_uuid(apps, schema_editor):
    """Populate ProjectEventStatus sodar_uuid fields"""
    ProjectEventStatus = apps.get_model('timeline', 'ProjectEventStatus')
    for status in ProjectEventStatus.objects.all():
        status.sodar_uuid = uuid.uuid4()
        status.save(update_fields=['sodar_uuid'])


def populate_object_ref_uuid(apps, schema_editor):
    """Populate TimelineEventObjectRef sodar_uuid fields"""
    TimelineEventObjectRef = apps.get_model(
        'timeline', 'TimelineEventObjectRef'
    )
    for status in TimelineEventObjectRef.objects.all():
        status.sodar_uuid = uuid.uuid4()
        status.save(update_fields=['sodar_uuid'])


# Functions from the following migrations need manual copying.
# Move them and any dependencies into this file, then update the
# RunPython operations to refer to the local versions:
# timeline.migrations.0010_populate_uuid
# timeline.migrations.0014_populate_uuid


class Migration(migrations.Migration):

    replaces = [
        ('timeline', '0001_initial'),
        ('timeline', '0002_projectevent_user'),
        ('timeline', '0003_rename_uuid'),
        ('timeline', '0004_update_uuid'),
        ('timeline', '0005_update_jsonfield'),
        ('timeline', '0006_update_user_optional'),
        ('timeline', '0007_allow_null_project'),
        ('timeline', '0008_projectevent_plugin'),
        ('timeline', '0009_create_status_uuid'),
        ('timeline', '0010_populate_uuid'),
        ('timeline', '0011_make_uuid_unique'),
        ('timeline', '0012_rename_models'),
        ('timeline', '0013_timelineeventobjectref_sodar_uuid'),
        ('timeline', '0014_populate_uuid'),
        ('timeline', '0015_make_uuid_unique'),
    ]

    initial = True

    dependencies = [
        ('projectroles', '0028_populate_finder_role'),
        ('projectroles', '0001_initial'),
        ('projectroles', '0019_project_public_guest_access'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectEvent',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'app',
                    models.CharField(
                        help_text='App from which the event was triggered',
                        max_length=255,
                    ),
                ),
                (
                    'event_name',
                    models.CharField(
                        help_text='Event ID string', max_length=255
                    ),
                ),
                (
                    'description',
                    models.TextField(
                        help_text='Description of status change (may include {object label} references)'
                    ),
                ),
                (
                    'extra_data',
                    models.JSONField(
                        default=dict, help_text='Additional event data as JSON'
                    ),
                ),
                (
                    'classified',
                    models.BooleanField(
                        default=False,
                        help_text='Event is classified (only viewable by user levels specified in rules)',
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='Event SODAR UUID',
                        unique=True,
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        help_text='Project to which the event belongs (null for no project)',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='events',
                        to='projectroles.project',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='User who initiated the event (optional)',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'plugin',
                    models.CharField(
                        blank=True,
                        help_text='Plugin to which the event is related (optional, if unset plugin with the name of the app is assumed)',
                        max_length=255,
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='ProjectEventObjectRef',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'label',
                    models.CharField(
                        help_text='Label for the object related to the event',
                        max_length=255,
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        help_text='Name or title of the object', max_length=255
                    ),
                ),
                (
                    'object_model',
                    models.CharField(
                        help_text='Object model as string', max_length=255
                    ),
                ),
                (
                    'object_uuid',
                    models.UUIDField(
                        blank=True, help_text='Object SODAR UUID', null=True
                    ),
                ),
                (
                    'extra_data',
                    models.JSONField(
                        default=dict,
                        help_text='Additional data related to the object as JSON',
                    ),
                ),
                (
                    'event',
                    models.ForeignKey(
                        help_text='Event to which the object belongs',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='event_objects',
                        to='timeline.projectevent',
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='ProjectEventStatus',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'timestamp',
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text='DateTime of the status change',
                    ),
                ),
                (
                    'status_type',
                    models.CharField(
                        help_text='Type of the status change', max_length=64
                    ),
                ),
                (
                    'description',
                    models.TextField(
                        blank=True,
                        help_text='Description of status change (optional)',
                    ),
                ),
                (
                    'extra_data',
                    models.JSONField(
                        default=dict, help_text='Additional status data as JSON'
                    ),
                ),
                (
                    'event',
                    models.ForeignKey(
                        help_text='Event to which the status change belongs',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='status_changes',
                        to='timeline.projectevent',
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='Status SODAR UUID',
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.RunPython(
            code=populate_event_status_uuid,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AlterField(
            model_name='projecteventstatus',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4, help_text='Status SODAR UUID', unique=True
            ),
        ),
        migrations.RenameModel(
            old_name='ProjectEvent',
            new_name='TimelineEvent',
        ),
        migrations.RenameModel(
            old_name='ProjectEventObjectRef',
            new_name='TimelineEventObjectRef',
        ),
        migrations.RenameModel(
            old_name='ProjectEventStatus',
            new_name='TimelineEventStatus',
        ),
        migrations.AddField(
            model_name='timelineeventobjectref',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text='Object reference SODAR UUID',
                null=True,
            ),
        ),
        migrations.RunPython(
            code=populate_object_ref_uuid,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AlterField(
            model_name='timelineeventobjectref',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text='Object reference SODAR UUID',
                unique=True,
            ),
        ),
    ]



from django.contrib import admin
from .models import Project, Role, RoleAssignment, AppSetting, ProjectInvite


admin.site.register(Project)
admin.site.register(Role)
admin.site.register(RoleAssignment)
admin.site.register(AppSetting)
admin.site.register(ProjectInvite)



"""Plugin tests for the adminalerts app"""

from test_plus.test import TestCase

# Projectroles dependency
from projectroles.plugins import SiteAppPluginPoint

from adminalerts.urls import urlpatterns
from adminalerts.tests.test_models import AdminAlertMixin


PLUGIN_NAME = 'adminalerts'
PLUGIN_TITLE = 'Admin Alerts'
PLUGIN_URL_ID = 'adminalerts:list'


# NOTE: Setting up the plugin is done during migration


class TestPlugins(AdminAlertMixin, TestCase):
    """Test adminalerts plugin"""

    def setUp(self):
        # Create users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        # Create alert
        self.alert = self.make_alert(
            message='alert',
            user=self.superuser,
            description='description',
            active=True,
        )

    def test_plugin_retrieval(self):
        """Test retrieving the plugin from the database"""
        plugin = SiteAppPluginPoint.get_plugin(PLUGIN_NAME)
        self.assertIsNotNone(plugin)
        self.assertEqual(plugin.get_model().name, PLUGIN_NAME)
        self.assertEqual(plugin.name, PLUGIN_NAME)
        self.assertEqual(plugin.get_model().title, PLUGIN_TITLE)
        self.assertEqual(plugin.title, PLUGIN_TITLE)
        self.assertEqual(plugin.entry_point_url_id, PLUGIN_URL_ID)

    def test_plugin_urls(self):
        """Test plugin URLs to ensure they're the same as in the app config"""
        plugin = SiteAppPluginPoint.get_plugin(PLUGIN_NAME)
        self.assertEqual(plugin.urls, urlpatterns)

    def test_get_messages(self):
        """Test the get_messages() function"""
        plugin = SiteAppPluginPoint.get_plugin(PLUGIN_NAME)
        messages = plugin.get_messages()
        message = messages[0]
        self.assertEqual(len(messages), 1)
        self.assertIn(self.alert.message, message['content'])
        self.assertEqual(message['color'], 'info')
        self.assertEqual(message['dismissable'], False)
        self.assertEqual(message['require_auth'], True)



"""Ajax API views for the appalerts app"""

from rest_framework.response import Response

# Projectroles dependency
from projectroles.views_ajax import SODARBaseAjaxView

from appalerts.models import AppAlert


class AppAlertStatusAjaxView(SODARBaseAjaxView):
    """View to get app alert status for user"""

    permission_required = 'appalerts.view_alerts'

    def get(self, request, **kwargs):
        # HACK: Manually refuse access to anonymous as this view is an exception
        if not request.user or request.user.is_anonymous:
            return Response({'detail': 'Anonymous access denied'}, status=401)
        return Response(
            {
                'alerts': AppAlert.objects.filter(
                    user=request.user, active=True
                ).count()
            },
            status=200,
        )


class AppAlertDismissAjaxView(SODARBaseAjaxView):
    """View to handle app alert dismissal in UI"""

    permission_required = 'appalerts.view_alerts'

    def post(self, request, **kwargs):
        # HACK: Manually refuse access to anonymous as this view is an exception
        if not request.user or request.user.is_anonymous:
            return Response({'detail': 'Anonymous access denied'}, status=401)
        alerts = AppAlert.objects.filter(user=request.user, active=True)
        if kwargs.get('appalert'):
            alerts = alerts.filter(sodar_uuid=kwargs['appalert'])
        if not alerts:
            return Response({'detail': 'Not found'}, status=404)
        for alert in alerts:
            alert.active = False
            alert.save()
        return Response({'detail': 'OK'}, status=200)



"""Models for the filesfolders app"""

import uuid

from typing import Optional

from django.conf import settings
from django.db import models
from django.db.models import Q, QuerySet

from db_file_storage.model_utils import delete_file, delete_file_if_needed

# Projectroles dependency
from projectroles.models import Project


# Access Django user model
AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


# Local constants
FILESFOLDERS_FLAGS = {
    'IMPORTANT': {
        'icon': 'mdi:information',
        'label': 'Important',
        'color': 'warning',
        'text_classes': 'font-weight-bold',
    },
    'FLAG': {
        'icon': 'mdi:flag',
        'label': 'Flagged',
        'color': 'info',
        'text_classes': 'font-weight-bold',
    },
    'FLAG_HEART': {
        'icon': 'mdi:heart',
        'label': 'Flagged (Heart)',
        'color': 'danger',
        'text_classes': 'font-weight-bold',
    },
    'REVOKED': {
        'icon': 'mdi:undo',
        'label': 'Revoked',
        'color': 'secondary',
        'text_classes': 'text-strikethrough',
    },
    'SUPERSEDED': {
        'icon': 'mdi:history',
        'label': 'Superseded',
        'color': 'dark',
        'text_classes': 'text-muted',
    },
}

FLAG_CHOICES = [
    (k, FILESFOLDERS_FLAGS[k]['label']) for k in sorted(FILESFOLDERS_FLAGS)
]


# Base class -------------------------------------------------------------------


class FilesfoldersManager(models.Manager):
    """Manager for custom table-level BaseFilesfoldersClass queries"""

    def find(
        self, search_terms: list[str], keywords: Optional[dict] = None
    ) -> QuerySet:
        """
        Return files, folders and/or hyperlinks matching the query.

        :param search_terms: Search terms (list of strings)
        :param keywords: Optional search keywords as key/value pairs (dict)
        :return: QuerySet of BaseFilesfolderClass objects
        """
        objects = super().get_queryset().order_by('name')
        term_query = Q()
        for t in search_terms:
            term_query.add(Q(name__icontains=t), Q.OR)
            term_query.add(Q(description__icontains=t), Q.OR)
            try:
                uuid.UUID(t.replace('-', ''))
                term_query.add(Q(sodar_uuid=t), Q.OR)
            except ValueError:
                pass
        return objects.filter(term_query)


class BaseFilesfoldersClass(models.Model):
    """Abstract class for all filesfolders objects"""

    #: Folder name
    name = models.CharField(
        max_length=255, unique=False, help_text='Name for the object'
    )

    #: Project in which the folder belongs
    project = models.ForeignKey(
        Project,
        related_name='%(app_label)s_%(class)s_objects',
        help_text='Project in which the object belongs',
        on_delete=models.CASCADE,
    )

    #: Folder under which object exists (null if root folder)
    folder = models.ForeignKey(
        'Folder',
        related_name='%(app_label)s_%(class)s_children',
        null=True,
        blank=True,
        help_text='Folder under which object exists (null if root folder)',
        on_delete=models.CASCADE,
    )

    #: User who owns the object
    owner = models.ForeignKey(
        AUTH_USER_MODEL,
        help_text='User who owns the object',
        on_delete=models.CASCADE,
    )

    #: DateTime of last modification
    date_modified = models.DateTimeField(
        auto_now=True, help_text='DateTime of last modification'
    )

    #: Flag for highlighting the item (optional)
    flag = models.CharField(
        max_length=64,
        unique=False,
        blank=True,
        null=True,
        choices=FLAG_CHOICES,
        help_text='Flag for highlighting the item (optional)',
    )

    #: Description (optional)
    description = models.CharField(
        max_length=255,
        unique=False,
        blank=True,
        help_text='Description (optional)',
    )

    #: Filesfolders SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Filesfolders SODAR UUID'
    )

    # Set manager for custom queries
    objects = FilesfoldersManager()

    class Meta:
        abstract = True


# Folder -----------------------------------------------------------------------


class Folder(BaseFilesfoldersClass):
    """Folder which stores filefolders objects"""

    class Meta:
        ordering = ['project', 'name']
        unique_together = ('project', 'folder', 'name')

    def __str__(self):
        return '{}: {}{}'.format(
            self.project.title,
            self.folder.get_path() if self.folder else 'root/',
            self.name,
        )

    def __repr__(self):
        values = (
            self.project.title,
            self.name,
            self.folder if self.folder else '/',
        )
        return 'Folder({})'.format(', '.join(repr(v) for v in values))

    def get_path(self) -> str:
        """Return full path as str"""
        if self.folder:
            ret = self.folder.get_path()
        else:
            ret = 'root/'
        ret += f'{self.name}/'
        return ret

    def is_empty(self) -> bool:
        """Return True if the folder contains no subfolders, files or links"""
        return (
            self.filesfolders_folder_children.count() == 0
            and self.filesfolders_file_children.count() == 0
            and self.filesfolders_hyperlink_children.count() == 0
        )

    def has_in_path(self, folder: 'Folder') -> bool:
        """Return True if folder exists in this folder's parent path"""
        if self.folder == folder:
            return True
        elif self.folder:
            return self.folder.has_in_path(folder)
        return False


# File -------------------------------------------------------------------------


class FileData(models.Model):
    """Class for storing actual file data in the Postgres database, needed by
    django-db-file-storage"""

    #: File data
    bytes = models.TextField()

    #: File name
    file_name = models.CharField(max_length=255)

    # Content type
    content_type = models.CharField(max_length=255)


class FileManager(FilesfoldersManager):
    """Manager for custom table-level File queries"""

    def get_folder_readme(
        self,
        project_pk: int,
        folder_pk: Optional[int],
        mimetype: str = 'text/markdown',
    ) -> Optional['File']:
        """
        Return the readme file for a folder or None if not found
        :param project_pk: Pk of the Project
        :param folder_pk: Pk of the Folder or None if root
        :param mimetype: Mimetype of the readme (default=text/markdown)
        :return: File or None
        """
        # NOTE: Can't just use get() with file__file__mimetype here
        readme_files = File.objects.filter(
            name__istartswith='readme.', project=project_pk, folder=folder_pk
        )
        for f in readme_files:
            if f.file.file.mimetype == mimetype:
                return f
        return None


class File(BaseFilesfoldersClass):
    """Small file uploaded using the filesfolders app"""

    #: Uploaded file using django-db-file-storage
    file = models.FileField(
        blank=True,
        null=True,
        upload_to='filesfolders.FileData/bytes/file_name/content_type',
        help_text='Uploaded file',
    )

    #: Allow providing a public URL for the file
    public_url = models.BooleanField(
        default=False, help_text='Allow providing a public URL for the file'
    )

    #: Secret string for creating public URL (auto created by form / serializer)
    secret = models.CharField(
        max_length=255,
        unique=True,
        blank=False,
        null=False,
        help_text='Secret string for creating public URL',
    )

    # Set manager for custom queries
    objects = FileManager()

    class Meta:
        ordering = ['folder', 'name']
        unique_together = ('project', 'folder', 'name')

    def __str__(self):
        return '{}: {}{}'.format(
            self.project.title,
            self.folder.get_path() if self.folder else 'root/',
            self.name,
        )

    def __repr__(self):
        values = (
            self.project.title,
            self.name,
            self.folder if self.folder else '/',
        )
        return 'File({})'.format(', '.join(repr(v) for v in values))

    def save(self, *args, **kwargs):
        """Override save for deleting file from database if needed"""
        delete_file_if_needed(self, 'file')
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """Override delete for deleting file from database"""
        super().delete(*args, **kwargs)
        delete_file(self, 'file')


class HyperLink(BaseFilesfoldersClass):
    """Hyperlink saved using the filesfolders app"""

    #: URL for the link
    url = models.URLField(
        max_length=2000, blank=False, null=False, help_text='URL for the link'
    )

    class Meta:
        ordering = ['folder', 'name']
        unique_together = ('project', 'folder', 'name')

    def __str__(self):
        return '{}: {}{}'.format(
            self.project.title,
            self.folder.name + ' / ' if self.folder else '',
            self.name,
        )

    def __repr__(self):
        values = (
            self.project.title,
            self.name,
            self.folder if self.folder else '',
        )
        return 'HyperLink({})'.format(', '.join(repr(v) for v in values))



"""Geticons management command for retrieving icons for iconify"""

import os

import urllib.request

from django.conf import settings
from django.core.management.base import BaseCommand

from projectroles.management.logging import ManagementCommandLogger


logger = ManagementCommandLogger(__name__)


# Local constants
JSON_URL = (
    'https://raw.githubusercontent.com/iconify/collections-json/'
    'master/collections.json'
)
COLL_URL = (
    'https://raw.githubusercontent.com/iconify/collections-json/'
    'master/json/{id}.json'
)


class Command(BaseCommand):
    help = 'Retrieves or updates JSON Iconify icons'

    def _download(self, url: str, base_path: str, file_name: str):
        """
        Download file.

        :param url: URL to file
        :param base_path: Base path in local file system
        :param file_name: File name
        """
        logger.debug(f'Downloading "{file_name}"..')
        response = urllib.request.urlopen(url)
        with open(os.path.join(base_path, file_name), 'wb') as f:
            f.write(response.read())
        logger.info(f'Download of "{file_name}" OK')

    def add_arguments(self, parser):
        parser.add_argument(
            '-c',
            '--collections',
            dest='collections',
            nargs='+',
            required=False,
            help='Additional collection IDs to retrieve',
        )
        parser.add_argument(
            '-p',
            '--path',
            dest='path',
            type=str,
            required=False,
            help='Base path if not {SITE_PACKAGE}/static/iconify',
        )

    def handle(self, *args, **options):
        logger.info('Retrieving/updating JSON icons..')
        colls = options.get('collections')
        if options.get('path'):
            iconify_path = options['path']
        else:
            iconify_path = os.path.join(settings.APPS_DIR, 'static', 'iconify')
        coll_path = os.path.join(iconify_path, 'json')

        # Create Iconify dirs if not there
        if not os.path.exists(coll_path):
            os.makedirs(coll_path, mode=0o755)
            logger.info(
                f'Created Iconify JSON directories in {settings.SITE_PACKAGE}'
                f'/static'
            )
        else:
            logger.debug('Found existing Iconify JSON dirs')

        # Make sure we have .gitkeep files so large JSONs are not committed
        open(os.path.join(iconify_path, '.gitkeep'), 'a').close()
        open(os.path.join(coll_path, '.gitkeep'), 'a').close()

        # Download collections.json
        self._download(JSON_URL, iconify_path, 'collections.json')
        # Download mdi.json
        self._download(
            COLL_URL.format(id='mdi'),
            coll_path,
            'mdi.json',
        )
        # Download extra collections if set
        if colls:
            logger.info(
                'Downloading {} additional collection{}..'.format(
                    len(colls), 's' if len(colls) != 1 else ''
                )
            )
            for c in colls:
                self._download(
                    COLL_URL.format(id=c),
                    coll_path,
                    c + '.json',
                )
        logger.info('Retrieval/update done')
        logger.info('Remember to run "collectstatic" after updating icons!')



"""Tests for views in the userprofile Django app"""

import uuid

from django.contrib import auth
from django.contrib.messages import get_messages
from django.core import mail
from django.forms.models import model_to_dict
from django.test import override_settings
from django.urls import reverse

from test_plus.test import TestCase

# Timeline dependency
from timeline.models import TimelineEvent

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.models import (
    AppSetting,
    SODARUserAdditionalEmail,
    SODAR_CONSTANTS,
)
from projectroles.tests.test_models import (
    EXAMPLE_APP_NAME,
    AppSettingMixin,
    SODARUserAdditionalEmailMixin,
)
from projectroles.views import FORM_INVALID_MSG
from projectroles.utils import build_secret

from userprofile.views import (
    SETTING_UPDATE_MSG,
    EMAIL_NOT_FOUND_MSG,
    EMAIL_ALREADY_VERIFIED_MSG,
    EMAIL_VERIFIED_MSG,
    EMAIL_VERIFY_RESEND_MSG,
)


app_settings = AppSettingAPI()
User = auth.get_user_model()


# SODAR constants
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']

# Local constants
INVALID_VALUE = 'INVALID VALUE'
ADD_EMAIL = 'add1@example.com'
ADD_EMAIL2 = 'add2@example.com'
ADD_EMAIL_SECRET = build_secret(32)


class UserViewTestBase(TestCase):
    """Base class for view testing"""

    def setUp(self):
        # Init superuser
        self.user = self.make_user('superuser')
        self.user.is_staff = True
        self.user.is_superuser = True
        self.user.save()


# View tests -------------------------------------------------------------------


class TestUserDetailView(SODARUserAdditionalEmailMixin, UserViewTestBase):
    """Tests for UserDetailView"""

    def test_get(self):
        """Test UserDetailView GET"""
        with self.login(self.user):
            response = self.client.get(reverse('userprofile:detail'))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context['user_settings'])
        self.assertEqual(response.context['add_emails'].count(), 0)
        self.assertEqual(response.context['site_read_only'], False)
        self.assertEqual(response.context['send_email'], True)
        self.assertEqual(response.context['site_mode'], SITE_MODE_SOURCE)

    def test_get_additional_email(self):
        """Test GET with additional email"""
        self.make_email(self.user, 'add@example.com')
        self.make_email(self.user, 'add_unverified@example.com', verified=False)
        with self.login(self.user):
            response = self.client.get(reverse('userprofile:detail'))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context['user_settings'])
        self.assertEqual(response.context['add_emails'].count(), 2)


class TestUserAppSettingsView(AppSettingMixin, UserViewTestBase):
    """Tests for UserAppSettingsView"""

    def _get_setting(self, name) -> AppSetting:
        return app_settings.get(EXAMPLE_APP_NAME, name, user=self.user)

    def setUp(self):
        super().setUp()
        # Init test setting
        self.setting_str = self.make_setting(
            plugin_name=EXAMPLE_APP_NAME,
            name='user_str_setting',
            setting_type=APP_SETTING_TYPE_STRING,
            value='test',
            user=self.user,
        )
        # Init integer setting
        self.setting_int = self.make_setting(
            plugin_name=EXAMPLE_APP_NAME,
            name='user_int_setting',
            setting_type=APP_SETTING_TYPE_INTEGER,
            value=170,
            user=self.user,
        )
        # Init test setting with options
        self.setting_str_options = self.make_setting(
            plugin_name=EXAMPLE_APP_NAME,
            name='user_str_setting_options',
            setting_type=APP_SETTING_TYPE_STRING,
            value='string1',
            user=self.user,
        )
        # Init integer setting with options
        self.setting_int_options = self.make_setting(
            plugin_name=EXAMPLE_APP_NAME,
            name='user_int_setting_options',
            setting_type=APP_SETTING_TYPE_INTEGER,
            value=0,
            user=self.user,
        )
        # Init boolean setting
        self.setting_bool = self.make_setting(
            plugin_name=EXAMPLE_APP_NAME,
            name='user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            user=self.user,
        )
        # Init json setting
        self.setting_json = self.make_setting(
            plugin_name=EXAMPLE_APP_NAME,
            name='user_json_setting',
            setting_type=APP_SETTING_TYPE_JSON,
            value=None,
            value_json={'Test': 'More'},
            user=self.user,
        )

    def test_get(self):
        """Test UserAppSettingsView GET"""
        with self.login(self.user):
            response = self.client.get(reverse('userprofile:settings_update'))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context['form'])
        field = response.context['form'].fields.get(
            'settings.example_project_app.user_str_setting'
        )
        self.assertIsNotNone(field)
        self.assertEqual(field.widget.attrs['placeholder'], 'Example string')
        field = response.context['form'].fields.get(
            'settings.example_project_app.user_int_setting'
        )
        self.assertIsNotNone(field)
        self.assertEqual(field.widget.attrs['placeholder'], 0)
        self.assertIsNotNone(
            response.context['form'].fields.get(
                'settings.example_project_app.user_str_setting_options'
            )
        )
        self.assertIsNotNone(
            response.context['form'].fields.get(
                'settings.example_project_app.user_int_setting_options'
            )
        )
        self.assertIsNotNone(
            response.context['form'].fields.get(
                'settings.example_project_app.user_bool_setting'
            )
        )
        self.assertIsNotNone(
            response.context['form'].fields.get(
                'settings.example_project_app.user_json_setting'
            )
        )

    def test_post(self):
        """Test POST"""
        self.assertEqual(self._get_setting('user_str_setting'), 'test')
        self.assertEqual(self._get_setting('user_int_setting'), 170)
        self.assertEqual(
            self._get_setting('user_str_setting_options'), 'string1'
        )
        self.assertEqual(self._get_setting('user_int_setting_options'), 0)
        self.assertEqual(self._get_setting('user_bool_setting'), True)
        self.assertEqual(
            self._get_setting('user_json_setting'), {'Test': 'More'}
        )

        values = {
            'settings.example_project_app.user_str_setting': 'another-text',
            'settings.example_project_app.user_int_setting': '123',
            'settings.example_project_app.user_str_setting_options': 'string2',
            'settings.example_project_app.user_int_setting_options': 1,
            'settings.example_project_app.user_bool_setting': False,
            'settings.example_project_app.'
            'user_json_setting': '{"Test": "Less"}',
            'settings.example_project_app.user_callable_setting': 'Test',
            'settings.example_project_app.user_callable_setting_options': str(
                self.user.sodar_uuid
            ),
            'settings.projectroles.project_list_highlight': False,
            'settings.projectroles.project_list_pagination': 10,
        }
        with self.login(self.user):
            response = self.client.post(
                reverse('userprofile:settings_update'), values
            )

        # Assert redirect
        with self.login(self.user):
            self.assertRedirects(response, reverse('userprofile:detail'))
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            SETTING_UPDATE_MSG,
        )
        # Assert settings state after update
        self.assertEqual(self._get_setting('user_str_setting'), 'another-text')
        self.assertEqual(self._get_setting('user_int_setting'), 123)
        self.assertEqual(
            self._get_setting('user_str_setting_options'), 'string2'
        )
        self.assertEqual(self._get_setting('user_int_setting_options'), 1)
        self.assertEqual(self._get_setting('user_bool_setting'), False)
        self.assertEqual(
            self._get_setting('user_json_setting'), {'Test': 'Less'}
        )
        self.assertEqual(self._get_setting('user_callable_setting'), 'Test')
        self.assertEqual(
            self._get_setting('user_callable_setting_options'),
            str(self.user.sodar_uuid),
        )

    def test_post_custom_validation(self):
        """Test POST with custom validation and invalid value"""
        values = {
            'settings.example_project_app.' 'user_str_setting': INVALID_VALUE,
            'settings.example_project_app.user_int_setting': '170',
            'settings.example_project_app.user_str_setting_options': 'string1',
            'settings.example_project_app.user_int_setting_options': '0',
            'settings.example_project_app.user_bool_setting': True,
            'settings.example_project_app.'
            'user_json_setting': '{"Test": "More"}',
            'settings.example_project_app.user_callable_setting': 'Test',
            'settings.example_project_app.user_callable_setting_options': str(
                self.user.sodar_uuid
            ),
            'settings.projectroles.project_list_highlight': False,
            'settings.projectroles.project_list_pagination': 10,
        }
        with self.login(self.user):
            response = self.client.post(
                reverse('userprofile:settings_update'), values
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            FORM_INVALID_MSG,
        )
        self.assertEqual(self._get_setting('user_str_setting'), 'test')


class TestUserEmailCreateView(SODARUserAdditionalEmailMixin, UserViewTestBase):
    """Tests for UserEmailCreateView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('userprofile:email_create')
        self.url_redirect = reverse('userprofile:detail')

    def test_get(self):
        """Test UserEmailCreateView GET"""
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context['form'])

    @override_settings(PROJECTROLES_SEND_EMAIL=False)
    def test_get_email_disabled(self):
        """Test GET with disabled email sending"""
        with self.login(self.user):
            response = self.client.get(self.url)
            self.assertRedirects(response, self.url_redirect)

    def test_post(self):
        """Test POST"""
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 0)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='email_create').count(), 0
        )
        self.assertEqual(len(mail.outbox), 0)
        with self.login(self.user):
            data = {
                'user': self.user.pk,
                'email': ADD_EMAIL,
                'secret': ADD_EMAIL_SECRET,
            }
            response = self.client.post(self.url, data)
            self.assertRedirects(response, self.url_redirect)
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 1)
        email = SODARUserAdditionalEmail.objects.first()
        expected = {
            'id': email.pk,
            'user': self.user.pk,
            'email': ADD_EMAIL,
            'secret': ADD_EMAIL_SECRET,
            'verified': False,
            'sodar_uuid': email.sodar_uuid,
        }
        self.assertEqual(model_to_dict(email), expected)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='email_create').count(), 1
        )
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].recipients(), [ADD_EMAIL])
        verify_url = reverse(
            'userprofile:email_verify', kwargs={'secret': email.secret}
        )
        self.assertIn(verify_url, mail.outbox[0].body)

    def test_post_existing_primary(self):
        """Test POST with email used as primary email for user"""
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 0)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='email_create').count(), 0
        )
        self.assertEqual(len(mail.outbox), 0)
        with self.login(self.user):
            data = {
                'user': self.user.pk,
                'email': self.user.email,
                'secret': ADD_EMAIL_SECRET,
            }
            response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 0)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='email_create').count(), 0
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_post_existing_additional(self):
        """Test POST with email used as additional email for user"""
        self.make_email(self.user, ADD_EMAIL)
        with self.login(self.user):
            data = {
                'user': self.user.pk,
                'email': ADD_EMAIL,
                'secret': ADD_EMAIL_SECRET,
            }
            response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 200)

    def test_post_multiple(self):
        """Test POST with different existing additional email"""
        self.make_email(self.user, ADD_EMAIL)
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 1)
        self.assertEqual(len(mail.outbox), 0)
        with self.login(self.user):
            data = {
                'user': self.user.pk,
                'email': ADD_EMAIL2,
                'secret': ADD_EMAIL_SECRET,
            }
            response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 2)
        self.assertEqual(len(mail.outbox), 1)


class TestUserEmailVerifyView(SODARUserAdditionalEmailMixin, UserViewTestBase):
    """Tests for UserEmailVerifyView"""

    def setUp(self):
        super().setUp()
        self.url_redirect = reverse('userprofile:detail')
        self.email = self.make_email(self.user, ADD_EMAIL, verified=False)
        self.url = reverse(
            'userprofile:email_verify', kwargs={'secret': self.email.secret}
        )

    def test_get(self):
        """Test UserEmailVerifyView GET"""
        self.assertEqual(self.email.verified, False)
        with self.login(self.user):
            response = self.client.get(self.url)
            self.assertRedirects(response, self.url_redirect)
        self.email.refresh_from_db()
        self.assertEqual(self.email.verified, True)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_VERIFIED_MSG.format(email=self.email.email),
        )

    def test_get_invalid_secret(self):
        """Test GET with invalid secret"""
        self.assertEqual(self.email.verified, False)
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'userprofile:email_verify',
                    kwargs={'secret': build_secret(32)},
                )
            )
        self.assertEqual(response.status_code, 302)
        self.email.refresh_from_db()
        self.assertEqual(self.email.verified, False)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_NOT_FOUND_MSG,
        )

    def test_get_wrong_user(self):
        """Test GET with wrong user"""
        user_new = self.make_user('user_new')
        self.assertEqual(self.email.verified, False)
        with self.login(user_new):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 302)
        self.email.refresh_from_db()
        self.assertEqual(self.email.verified, False)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_NOT_FOUND_MSG,
        )

    def test_get_verified(self):
        """Test GET with verified email"""
        self.email.verified = True
        self.email.save()
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 302)
        self.email.refresh_from_db()
        self.assertEqual(self.email.verified, True)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_ALREADY_VERIFIED_MSG,
        )


class TestUserEmailVerifyResendView(
    SODARUserAdditionalEmailMixin, UserViewTestBase
):
    """Tests for UserEmailVerifyResendView"""

    def setUp(self):
        super().setUp()
        self.url_redirect = reverse('userprofile:detail')
        self.email = self.make_email(self.user, ADD_EMAIL, verified=False)
        self.url = reverse(
            'userprofile:email_verify_resend',
            kwargs={'sodaruseradditionalemail': self.email.sodar_uuid},
        )

    def test_get(self):
        """Test UserEmailVerifyResendView GET"""
        self.assertEqual(len(mail.outbox), 0)
        with self.login(self.user):
            response = self.client.get(self.url)
            self.assertRedirects(response, self.url_redirect)
        self.email.refresh_from_db()
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_VERIFY_RESEND_MSG.format(email=self.email.email),
        )
        self.assertEqual(len(mail.outbox), 1)

    def test_get_invalid_uuid(self):
        """Test GET with invalid UUID"""
        self.assertEqual(self.email.verified, False)
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'userprofile:email_verify_resend',
                    kwargs={'sodaruseradditionalemail': uuid.uuid4()},
                )
            )
        self.assertEqual(response.status_code, 302)
        self.email.refresh_from_db()
        self.assertEqual(self.email.verified, False)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_NOT_FOUND_MSG,
        )

    def test_get_wrong_user(self):
        """Test GET with wrong user"""
        user_new = self.make_user('user_new')
        self.assertEqual(self.email.verified, False)
        with self.login(user_new):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 302)
        self.email.refresh_from_db()
        self.assertEqual(self.email.verified, False)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_NOT_FOUND_MSG,
        )

    def test_get_verified(self):
        """Test GET with verified email"""
        self.email.verified = True
        self.email.save()
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 302)
        self.email.refresh_from_db()
        self.assertEqual(self.email.verified, True)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            EMAIL_ALREADY_VERIFIED_MSG,
        )


class TestUserEmailDeleteView(SODARUserAdditionalEmailMixin, UserViewTestBase):
    """Tests for UserEmailDeleteView"""

    def setUp(self):
        super().setUp()
        self.url_redirect = reverse('userprofile:detail')
        self.email = self.make_email(self.user, ADD_EMAIL, verified=False)
        self.url = reverse(
            'userprofile:email_delete',
            kwargs={'sodaruseradditionalemail': self.email.sodar_uuid},
        )

    def test_get(self):
        """Test UserEmailDeleteView GET"""
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['object'], self.email)

    def test_post(self):
        """Test POST"""
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 1)
        with self.login(self.user):
            response = self.client.post(self.url)
            self.assertRedirects(response, self.url_redirect)
        self.assertEqual(SODARUserAdditionalEmail.objects.count(), 0)



"""Plugins for the siteinfo app"""

# Projectroles dependency
from projectroles.plugins import SiteAppPluginPoint

from siteinfo.urls import urlpatterns


class SiteAppPlugin(SiteAppPluginPoint):
    """Projectroles plugin for registering the app"""

    #: Name (used as plugin ID)
    name = 'siteinfo'

    #: Title (used in templates)
    title = 'Site Info'

    #: UI URLs
    urls = urlpatterns

    #: Iconify icon
    icon = 'mdi:bar-chart'

    #: Description string
    description = 'Site information and app statistics'

    #: Entry point URL ID
    entry_point_url_id = 'siteinfo:info'

    #: Required permission for displaying the app
    app_permission = 'siteinfo:view_info'



"""Plugins for the filesfolders app"""

from typing import Optional, Union
from uuid import UUID

from django.conf import settings
from django.urls import reverse

# Projectroles dependency
from projectroles.models import (
    Project,
    SODARUser,
    SODAR_CONSTANTS,
    CAT_DELIMITER,
)
from projectroles.plugins import (
    ProjectAppPluginPoint,
    PluginAppSettingDef,
    PluginObjectLink,
    PluginSearchResult,
    PluginCategoryStatistic,
)
from projectroles.utils import get_display_name

from filesfolders.models import File, Folder, HyperLink
from filesfolders.urls import urlpatterns


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']

# Local constants
SHOW_LIST_COLUMNS = getattr(settings, 'FILESFOLDERS_SHOW_LIST_COLUMNS', False)


class ProjectAppPlugin(ProjectAppPluginPoint):
    """Plugin for registering app with Projectroles"""

    # Properties required by django-plugins ------------------------------

    #: Name (used as plugin ID)
    name = 'filesfolders'

    #: Title (used in templates)
    title = 'Files'

    #: UI URLs
    urls = urlpatterns

    # Properties defined in ProjectAppPluginPoint -----------------------

    #: App setting definitions
    app_settings = [
        PluginAppSettingDef(
            name='allow_public_links',
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_BOOLEAN,
            default=False,
            label='Allow public links',
            description='Allow generation of public links for files',
        )
    ]

    #: Iconify icon
    icon = 'mdi:file'

    #: Entry point URL ID (must take project sodar_uuid as "project" argument)
    entry_point_url_id = 'filesfolders:list'

    #: Description string
    description = (
        'Smaller files (e.g., reports, spreadsheets, and presentations)'
    )

    #: Required permission for accessing the app
    app_permission = 'filesfolders.view_data'

    #: Enable or disable general search from project title bar
    search_enable = True

    #: List of search object types for the app
    search_types = ['file', 'folder', 'link']

    #: Search results template
    search_template = 'filesfolders/_search_results.html'

    #: App card template for the project details page
    details_template = 'filesfolders/_details_card.html'

    #: App card title for the project details page
    details_title = 'Files Overview'

    #: Position in plugin ordering
    plugin_ordering = 30

    project_list_columns = {
        'files': {
            'title': 'Files',
            'width': 75,
            'description': 'files stored for the project',
            'active': SHOW_LIST_COLUMNS,
            'ordering': 35,
            'align': 'right',
        },
        'links': {
            'title': 'Links',
            'width': 75,
            'description': 'Hyperlinks defined in the files app',
            'active': SHOW_LIST_COLUMNS,
            'ordering': 25,
            'align': 'right',
        },
    }

    #: Names of plugin specific Django settings to display in siteinfo
    info_settings = [
        'FILESFOLDERS_LINK_BAD_REQUEST_MSG',
        'FILESFOLDERS_MAX_ARCHIVE_SIZE',
        'FILESFOLDERS_MAX_UPLOAD_SIZE',
        'FILESFOLDERS_SERVE_AS_ATTACHMENT',
        'FILESFOLDERS_SHOW_LIST_COLUMNS',
    ]

    def get_object_link(
        self, model_str: str, uuid: Union[str, UUID]
    ) -> Optional[PluginObjectLink]:
        """
        Return URL referring to an object used by the app, along with a name to
        be shown to the user for linking.

        :param model_str: Object class (string)
        :param uuid: sodar_uuid of the referred object
        :return: PluginObjectLink or None if not found
        """
        obj = self.get_object(eval(model_str), uuid)
        if not obj:
            return None
        elif obj.__class__ == File:
            return PluginObjectLink(
                url=reverse(
                    'filesfolders:file_serve',
                    kwargs={'file': obj.sodar_uuid, 'file_name': obj.name},
                ),
                name=obj.name,
                blank=True,
            )
        elif obj.__class__ == Folder:
            return PluginObjectLink(
                url=reverse(
                    'filesfolders:list', kwargs={'folder': obj.sodar_uuid}
                ),
                name=obj.name,
            )
        elif obj.__class__ == HyperLink:
            return PluginObjectLink(url=obj.url, name=obj.name, blank=True)
        return None

    def search(
        self,
        search_terms: list[str],
        user: SODARUser,
        search_type: Optional[str] = None,
        keywords: Optional[list[str]] = None,
    ) -> list[PluginSearchResult]:
        """
        Return app items based on one or more search terms, user, optional type
        and optional keywords.

        :param search_terms: Search terms to be joined with the OR operator
                             (list of strings)
        :param user: User object for user initiating the search
        :param search_type: String
        :param keywords: List (optional)
        :return: List of PluginSearchResult objects
        """
        items = []
        if not search_type:
            files = File.objects.find(search_terms, keywords)
            folders = Folder.objects.find(search_terms, keywords)
            links = HyperLink.objects.find(search_terms, keywords)
            items = list(files) + list(folders) + list(links)
            items.sort(key=lambda x: x.name.lower())
        elif search_type == 'file':
            items = File.objects.find(search_terms, keywords).order_by('name')
        elif search_type == 'folder':
            items = Folder.objects.find(search_terms, keywords).order_by('name')
        elif search_type == 'link':
            items = HyperLink.objects.find(search_terms, keywords).order_by(
                'name'
            )
        if items:
            items = [
                x
                for x in items
                if user.has_perm('filesfolders.view_data', x.project)
            ]
        ret = PluginSearchResult(
            category='all',
            title='Files, Folders and Links',
            search_types=['file', 'folder', 'link'],
            items=items,
        )
        return [ret]

    def get_statistics(self) -> dict:
        return {
            'file_count': {
                'label': 'Files',
                'value': File.objects.all().count(),
            },
            'folder_count': {
                'label': 'Folders',
                'value': Folder.objects.all().count(),
            },
            'link_count': {
                'label': 'Hyperlinks',
                'value': HyperLink.objects.all().count(),
            },
        }

    def get_category_stats(
        self, category: Project
    ) -> list[PluginCategoryStatistic]:
        """
        Return app statistics for the given category. Expected to return
        cumulative statistics for all projects under the category and its
        possible subcategories.

        :param category: Project object of CATEGORY type
        :return: List of PluginCategoryStatistic objects
        """
        children = Project.objects.filter(
            type=PROJECT_TYPE_PROJECT,
            full_title__startswith=category.full_title + CAT_DELIMITER,
        )
        val = File.objects.filter(project__in=children).count()
        desc = 'Files uploaded to {} in this {}'.format(
            get_display_name(PROJECT_TYPE_PROJECT, plural=True),
            get_display_name(PROJECT_TYPE_CATEGORY),
        )
        return [
            PluginCategoryStatistic(
                plugin=self,
                title='Files',
                value=val,
                description=desc,
                icon='mdi:file',
            )
        ]

    def get_project_list_value(
        self, column_id: str, project: Project, user: SODARUser
    ) -> Union[str, int, None]:
        """
        Return a value for the optional additional project list column specific
        to a project.

        :param column_id: ID of the column (string)
        :param project: Project object
        :param user: User object (current user)
        :return: String (may contain HTML), integer or None
        """
        count = 0
        if column_id == 'files':
            count = File.objects.filter(project=project).count()
        elif column_id == 'links':
            count = HyperLink.objects.filter(project=project).count()
        if count > 0:
            url = reverse(
                'filesfolders:list', kwargs={'project': project.sodar_uuid}
            )
            return f'<a href="{url}">{count}</a>'
        return count



"""Logging helpers for management commands in SODAR Core based sites"""

import logging

from django.conf import settings


CONSOLE_HANDLER = 'console'


class ManagementCommandLogger:
    """Management command logger wrapper for SODAR Core based sites"""

    #: Python logger used by the wrapper
    logger = None

    #: Site logging level
    site_level = getattr(
        logging, getattr(settings, 'LOGGING_LEVEL', 'INFO'), 'INFO'
    )

    #: Whether console logging is enabled
    console_log = CONSOLE_HANDLER in settings.LOGGING.get('handlers', {}).keys()

    #: Disable console output if True
    disable_output = getattr(settings, 'LOGGING_DISABLE_CMD_OUTPUT', False)

    def __init__(self, name):
        self.logger = logging.getLogger(name)

    def log(self, message, level=logging.INFO):
        """
        Print and log output from e.g. a management command, where we always
        want to output a message but may not want to always log it.

        :param message: Output message (string)
        :param level: Log level (integer or string, default=logging.INFO)
        """
        if isinstance(level, str):
            level = getattr(logging, level)
        self.logger.log(level, message)
        if self.disable_output:
            return
        printed = False
        if level >= self.site_level:
            printed = self.console_log
        if not printed:
            print(message)

    def critical(self, message):
        """Log and print CRITICAL level message"""
        self.log(message, logging.CRITICAL)

    def error(self, message):
        """Log and print ERROR level message"""
        self.log(message, logging.ERROR)

    def warning(self, message):
        """Log and print WARNING level message"""
        self.log(message, logging.WARNING)

    def info(self, message):
        """Log and print INFO level message"""
        self.log(message, logging.INFO)

    def debug(self, message):
        """Log and print DEBUG level message"""
        self.log(message, logging.DEBUG)



"""Rule definitions for the bgjobs app"""

import rules

# Projectroles dependency
from projectroles import rules as pr_rules  # To access common predicates


# Predicates -------------------------------------------------------------------


# TODO: If we need to assign new predicates, we do it here


# Rules ------------------------------------------------------------------------


# TODO: Rules should not be needed, use permissions for user rights


# Permissions ------------------------------------------------------------------

# Allow viewing of background jobs
rules.add_perm(
    'bgjobs.view_data',
    pr_rules.is_project_owner
    | pr_rules.is_project_delegate
    | pr_rules.is_project_contributor
    | pr_rules.is_project_guest,
)

# Allow viewing of background jobs
rules.add_perm(
    'bgjobs.view_jobs_own',
    pr_rules.is_project_owner
    | pr_rules.is_project_delegate
    | pr_rules.is_project_contributor
    | pr_rules.is_project_guest,
)

# Allow creating background jobs
rules.add_perm(
    'bgjobs.create_bgjob',
    (
        pr_rules.is_project_owner
        | pr_rules.is_project_delegate
        | pr_rules.is_project_contributor
    )
    & pr_rules.is_site_writable,
)

# Allow modifying or deleting the user's background jobs
rules.add_perm(
    'bgjobs.update_bgjob_own',
    (
        pr_rules.is_project_owner
        | pr_rules.is_project_delegate
        | pr_rules.is_project_contributor
        | pr_rules.is_project_guest
    )
    & pr_rules.is_site_writable,
)

# Allow modifying or deleting all background jobs
rules.add_perm(
    'bgjobs.update_bgjob_all',
    (pr_rules.is_project_owner | pr_rules.is_project_delegate)
    & pr_rules.is_site_writable,
)

# Allow viewing site-wide background jobs (not project-specific)
rules.add_perm('bgjobs.site_view_data', rules.is_superuser)



import rules

# Projectroles dependency
from projectroles import rules as pr_rules  # To access common predicates


# Predicates -------------------------------------------------------------


@rules.predicate
def can_delete_email(user, obj):
    return obj.user == user


# Rules ------------------------------------------------------------------


# Rules should not be needed, use permissions for user rights


# Permissions ------------------------------------------------------------


# Allow viewing user details
rules.add_perm('userprofile.view_detail', rules.is_authenticated)

# Allow updating settings
rules.add_perm(
    'userprofile.update_settings',
    rules.is_authenticated & pr_rules.is_site_writable,
)

# Allow creating additional email
rules.add_perm(
    'userprofile.create_email',
    pr_rules.is_source_site
    & rules.is_authenticated
    & pr_rules.is_site_writable,
)

# Allow deleting additional email
rules.add_perm(
    'userprofile.delete_email',
    pr_rules.is_source_site & can_delete_email & pr_rules.is_site_writable,
)



import cProfile
import sys

from io import StringIO

from django.conf import settings
from django.utils.deprecation import MiddlewareMixin


class ProfilerMiddleware(MiddlewareMixin):
    """
    cProfile based profiling middleware.
    Based on: https://djangosnippets.org/snippets/727/
    """

    def process_view(self, request, callback, callback_args, callback_kwargs):
        if settings.DEBUG and 'prof' in request.GET:
            self.profiler = cProfile.Profile()
            args = (request,) + callback_args
            return self.profiler.runcall(callback, *args, **callback_kwargs)

    def process_response(self, request, response):
        if settings.DEBUG and 'prof' in request.GET:
            self.profiler.create_stats()
            out = StringIO()
            old_stdout, sys.stdout = sys.stdout, out
            self.profiler.print_stats(1)
            sys.stdout = old_stdout
            response.content = f'<pre>{out.getvalue()}</pre>'
        return response



"""Tests for REST API views in the sodarcache app"""

import json

from django.conf import settings
from django.forms.models import model_to_dict
from django.test import override_settings
from django.urls import reverse

from test_plus.test import APITestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
)
from projectroles.tests.test_views_api import SODARAPIViewTestMixin

from sodarcache.models import JSONCacheItem
from sodarcache.views_api import (
    SODARCACHE_API_MEDIA_TYPE,
    SODARCACHE_API_DEFAULT_VERSION,
)


plugin_api = PluginAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
TEST_APP_NAME = 'example_project_app'
ITEM_NAME = 'test_item'
DATA_KEY = 'test_key'
DATA_VAL = 'test_val'
DATA_VAL_UPDATED = 'test_val_updated'
BACKEND_PLUGINS_NO_CACHE = settings.ENABLED_BACKEND_PLUGINS.copy()
BACKEND_PLUGINS_NO_CACHE.remove('sodar_cache')


class SodarcacheAPIViewTestBase(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    SODARAPIViewTestMixin,
    APITestCase,
):
    """Base class for sodarcache REST API view testing"""

    media_type = SODARCACHE_API_MEDIA_TYPE
    api_version = SODARCACHE_API_DEFAULT_VERSION

    def setUp(self):
        super().setUp()
        # Get cache backend
        self.cache_backend = plugin_api.get_backend_api('sodar_cache')
        # Init roles
        self.init_roles()
        # Init users
        self.user = self.make_user('superuser')
        self.user.is_staff = True
        self.user.is_superuser = True
        self.user.save()
        self.user_owner = self.make_user('owner')
        # Init project
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, None
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        # Init cache item
        self.item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=TEST_APP_NAME,
            user=self.user_owner,
            name=ITEM_NAME,
            data={DATA_KEY: DATA_VAL},
        )
        # Get knox token for self.user
        self.knox_token = self.get_token(self.user)


class TestCacheItemRetrieveAPIView(SodarcacheAPIViewTestBase):
    """Tests for CacheItemRetrieveAPIView"""

    def test_get(self):
        """Test CacheItemRetrieveAPIView GET"""
        url = reverse(
            'sodarcache:api_retrieve',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 200)
        expected = {
            'project': str(self.project.sodar_uuid),
            'app_name': TEST_APP_NAME,
            'name': ITEM_NAME,
            'user': str(self.user_owner.sodar_uuid),
            'data': {DATA_KEY: DATA_VAL},
            'date_modified': self.get_drf_datetime(self.item.date_modified),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_get_no_user(self):
        """Test GET with no user for item"""
        self.item.user = None
        self.item.save()
        url = reverse(
            'sodarcache:api_retrieve',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)['user'], None)

    def test_get_no_item(self):
        """Test GET with non-existing item"""
        url = reverse(
            'sodarcache:api_retrieve',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': 'INVALID_ITEM_NAME',
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 404)

    def test_get_invalid_app_name(self):
        """Test GET with invalid app name"""
        url = reverse(
            'sodarcache:api_retrieve',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': 'INVALID_APP_NAME',
                'item_name': ITEM_NAME,
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 400)

    @override_settings(ENABLED_BACKEND_PLUGINS=BACKEND_PLUGINS_NO_CACHE)
    def test_get_backend_disabled(self):
        """Test GET with disabled sodarcache backend"""
        url = reverse(
            'sodarcache:api_retrieve',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 503)


class TestCacheItemDateRetrieveAPIView(SodarcacheAPIViewTestBase):
    """Tests for CacheItemDateRetrieveAPIView"""

    def test_get(self):
        """Test CacheItemDateRetrieveAPIView GET"""
        url = reverse(
            'sodarcache:api_retrieve_date',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        with self.login(self.user):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        expected = self.cache_backend.get_update_time(
            app_name=TEST_APP_NAME, name=ITEM_NAME
        )
        self.assertEqual(response.data['update_time'], expected)

    def test_get_no_item(self):
        """Test GET with non-existing item"""
        url = reverse(
            'sodarcache:api_retrieve_date',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': 'INVALID_ITEM_NAME',
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 404)

    def test_get_invalid_app_name(self):
        """Test GET with invalid app name"""
        url = reverse(
            'sodarcache:api_retrieve_date',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': 'INVALID_APP_NAME',
                'item_name': ITEM_NAME,
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 400)

    @override_settings(ENABLED_BACKEND_PLUGINS=BACKEND_PLUGINS_NO_CACHE)
    def test_get_backend_disabled(self):
        """Test GET with disabled sodarcache backend"""
        url = reverse(
            'sodarcache:api_retrieve_date',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        response = self.request_knox(url)
        self.assertEqual(response.status_code, 503)


class TestCacheItemSetAPIView(SodarcacheAPIViewTestBase):
    """Tests for CacheItemSetAPIView"""

    def test_post_create(self):
        """Test CacheItemSetAPIView POST to create item"""
        url = reverse(
            'sodarcache:api_set',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': 'new_test_item',
            },
        )
        post_data = {'data': json.dumps({DATA_KEY: DATA_VAL})}
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        response = self.request_knox(url, method='POST', data=post_data)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {'detail': 'ok'})
        self.assertEqual(JSONCacheItem.objects.all().count(), 2)
        item = JSONCacheItem.objects.get(name='new_test_item')
        expected = {
            'id': item.pk,
            'project': self.project.pk,
            'app_name': TEST_APP_NAME,
            'user': self.user.pk,
            'name': 'new_test_item',
            'data': json.dumps({DATA_KEY: DATA_VAL}),
            'sodar_uuid': item.sodar_uuid,
        }
        model_dict = model_to_dict(item)
        self.assertEqual(model_dict, expected)

    def test_post_update(self):
        """Test POST to update item"""
        url = reverse(
            'sodarcache:api_set',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        post_data = {'data': json.dumps({DATA_KEY: DATA_VAL_UPDATED})}
        response = self.request_knox(url, method='POST', data=post_data)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {'detail': 'ok'})
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        item = JSONCacheItem.objects.get(name=ITEM_NAME)
        expected = {
            'id': item.pk,
            'project': self.project.pk,
            'app_name': TEST_APP_NAME,
            'user': self.user.pk,
            'name': ITEM_NAME,
            'data': json.dumps({DATA_KEY: DATA_VAL_UPDATED}),
            'sodar_uuid': item.sodar_uuid,
        }
        model_dict = model_to_dict(item)
        self.assertEqual(model_dict, expected)

    @override_settings(ENABLED_BACKEND_PLUGINS=BACKEND_PLUGINS_NO_CACHE)
    def test_post_backend_disabled(self):
        """Test POST with disabled sodarcache backend"""
        url = reverse(
            'sodarcache:api_set',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        post_data = {'data': json.dumps({DATA_KEY: DATA_VAL_UPDATED})}
        response = self.request_knox(url, method='POST', data=post_data)
        self.assertEqual(response.status_code, 503)



"""Plugin tests for the filesfolders app"""

import uuid

from django.urls import reverse

from test_plus.test import TestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import ProjectAppPluginPoint, PluginObjectLink
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
)
from projectroles.utils import build_secret

from filesfolders.tests.test_models import (
    FolderMixin,
    FileMixin,
    HyperLinkMixin,
)
from filesfolders.urls import urlpatterns


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
PLUGIN_NAME = 'filesfolders'
PLUGIN_TITLE = 'Files'
PLUGIN_URL_ID = 'filesfolders:list'
SETTING_KEY = 'allow_public_links'


class TestPlugin(
    ProjectMixin,
    RoleMixin,
    FolderMixin,
    FileMixin,
    HyperLinkMixin,
    RoleAssignmentMixin,
    TestCase,
):
    """Tests for filesfolders project app plugin"""

    def setUp(self):
        # Init superuser
        self.user = self.make_user('superuser')
        self.user.is_staff = True
        self.user.is_superuser = True
        self.user.save()
        # Init roles
        self.init_roles()
        # Init category and project
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )
        # Init file
        self.file_content = bytes('content'.encode('utf-8'))
        self.file = self.make_file(
            name='file.txt',
            file_name='file.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.user,
            description='',
            public_url=True,
            secret=build_secret(),
        )
        # Init folder
        self.folder = self.make_folder(
            name='folder',
            project=self.project,
            folder=None,
            owner=self.user,
            description='',
        )
        # Init link
        self.hyperlink = self.make_hyperlink(
            name='Link',
            url='http://www.google.com/',
            project=self.project,
            folder=None,
            owner=self.user,
            description='',
        )
        # Get plugin
        self.plugin = ProjectAppPluginPoint.get_plugin(PLUGIN_NAME)

    def test_plugin_retrieval(self):
        """Test retrieving ProjectAppPlugin from the database"""
        self.assertIsNotNone(self.plugin)
        self.assertEqual(self.plugin.get_model().name, PLUGIN_NAME)
        self.assertEqual(self.plugin.name, PLUGIN_NAME)
        self.assertEqual(self.plugin.get_model().title, PLUGIN_TITLE)
        self.assertEqual(self.plugin.title, PLUGIN_TITLE)
        self.assertEqual(self.plugin.entry_point_url_id, PLUGIN_URL_ID)

    def test_plugin_urls(self):
        """Test plugin URLs to ensure they're the same as in the app config"""
        self.assertEqual(self.plugin.urls, urlpatterns)

    def test_get_object_link_file(self):
        """Test get_object_link() for a File object"""
        url = reverse(
            'filesfolders:file_serve',
            kwargs={'file': self.file.sodar_uuid, 'file_name': self.file.name},
        )
        ret = self.plugin.get_object_link('File', self.file.sodar_uuid)
        self.assertIsInstance(ret, PluginObjectLink)
        self.assertEqual(ret.url, url)
        self.assertEqual(ret.name, self.file.name)
        self.assertEqual(ret.blank, True)

    def test_get_object_link_folder(self):
        """Test get_object_link() for a Folder object"""
        url = reverse(
            'filesfolders:list', kwargs={'folder': self.folder.sodar_uuid}
        )
        ret = self.plugin.get_object_link('Folder', self.folder.sodar_uuid)
        self.assertEqual(ret.url, url)
        self.assertEqual(ret.name, self.folder.name)
        self.assertEqual(ret.blank, False)

    def test_get_object_link_hyperlink(self):
        """Test get_object_link() for a HyperLink object"""
        ret = self.plugin.get_object_link(
            'HyperLink', self.hyperlink.sodar_uuid
        )
        self.assertEqual(ret.url, self.hyperlink.url)
        self.assertEqual(ret.name, self.hyperlink.name)
        self.assertEqual(ret.blank, True)

    def test_get_object_link_fail(self):
        """Test get_object_link() with a non-existent object"""
        self.assertEqual(
            self.plugin.get_object_link('File', uuid.uuid4()), None
        )

    def test_get_category_stats(self):
        """Test get_category_stats()"""
        ret = self.plugin.get_category_stats(self.category)
        self.assertEqual(len(ret), 1)
        self.assertIsInstance(ret[0].plugin, self.plugin.__class__)
        self.assertEqual(ret[0].title, 'Files')
        self.assertEqual(ret[0].value, 1)

    def test_get_category_stats_no_files(self):
        """Test get_category_stats() with no files"""
        self.file.delete()
        ret = self.plugin.get_category_stats(self.category)
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0].value, 0)

    def test_get_category_stats_multi_project(self):
        """Test get_category_stats() with files in multiple projects"""
        new_project = self.make_project(
            'NewProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.file = self.make_file(
            name='file.txt',
            file_name='file.txt',
            file_content=self.file_content,
            project=new_project,
            folder=None,
            owner=self.user,
            description='',
            public_url=True,
            secret=build_secret(),
        )
        ret = self.plugin.get_category_stats(self.category)
        self.assertEqual(ret[0].value, 2)  # Both files should be counted

    def test_get_category_stats_multi_category(self):
        """Test get_category_stats() with files in multiple categories"""
        # Create new root category and place new project under it
        new_cat = self.make_project('NewCategory', PROJECT_TYPE_CATEGORY, None)
        new_project = self.make_project(
            'NewProject', PROJECT_TYPE_PROJECT, new_cat
        )
        self.file = self.make_file(
            name='file.txt',
            file_name='file.txt',
            file_content=self.file_content,
            project=new_project,
            folder=None,
            owner=self.user,
            description='',
            public_url=True,
            secret=build_secret(),
        )
        ret = self.plugin.get_category_stats(self.category)
        self.assertEqual(ret[0].value, 1)  # Only one file should be counted



from django.apps import AppConfig


class ProjectrolesConfig(AppConfig):
    name = 'projectroles'

    def ready(self):
        import projectroles.checks  # noqa
        import projectroles.signals  # noqa



"""REST API views for the sodarcache app"""

from typing import Any

from rest_framework import serializers
from rest_framework.exceptions import APIException, NotFound, ParseError
from rest_framework.generics import RetrieveAPIView
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.versioning import AcceptHeaderVersioning
from rest_framework.views import APIView

from drf_spectacular.utils import extend_schema, inline_serializer

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.views_api import SODARAPIGenericProjectMixin

from sodarcache.serializers import JSONCacheItemSerializer


plugin_api = PluginAPI()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME = 'sodarcache'
SODARCACHE_API_MEDIA_TYPE = (
    'application/vnd.bihealth.sodar-core.sodarcache+json'
)
SODARCACHE_API_DEFAULT_VERSION = '1.0'
SODARCACHE_API_ALLOWED_VERSIONS = ['1.0']


# Base Classes and Mixins ------------------------------------------------------


class SodarcacheAPIViewMixin:
    """
    Sodarcache API view versioning mixin for overriding media type and
    accepted versions, as well as providing helpers
    """

    class SodarcacheAPIRenderer(JSONRenderer):
        media_type = SODARCACHE_API_MEDIA_TYPE

    class SodarcacheAPIVersioning(AcceptHeaderVersioning):
        allowed_versions = SODARCACHE_API_ALLOWED_VERSIONS
        default_version = SODARCACHE_API_DEFAULT_VERSION

    class BackendUnavailable(APIException):
        status_code = 503
        default_detail = 'Cache backend not enabled'
        default_code = 'service_unavailable'

    renderer_classes = [SodarcacheAPIRenderer]
    versioning_class = SodarcacheAPIVersioning

    @classmethod
    def get_backend(cls) -> Any:
        """
        Return sodarcache backend or raise 503 if not enabled.

        :Return: SodarCacheAPI object
        """
        cache_backend = plugin_api.get_backend_api('sodar_cache')
        if not cache_backend:
            raise cls.BackendUnavailable()
        return cache_backend


# API Views --------------------------------------------------------------------


class CacheItemRetrieveAPIView(
    SodarcacheAPIViewMixin, SODARAPIGenericProjectMixin, RetrieveAPIView
):
    """
    Retrieve a cache item along with its data. Returns 404 if cache item is not
    set.

    **URL:** ``/cache/api/retrieve/{Project.sodar_uuid}/{app_name}/{item_name}``

    **Methods:** ``GET``

    **Returns:**

    - ``project``: Project UUID (string)
    - ``app_name``: Name of app to which the item belongs (string)
    - ``name``: Item name (string)
    - ``data``: Item data (JSON)
    - ``date_modified``: Item modification datetime (YYYY-MM-DDThh:mm:ssZ)
    - ``user``: UUID of user who created the item (string)

    **Version Changes:**

    - ``2.0``: Return ``user`` as UUID instead of ``SODARUserSerializer`` dict
    """

    permission_required = 'sodarcache.get_cache_value'
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = JSONCacheItemSerializer

    def get_object(self):
        """
        Override get_object() to return object by project, app name and name.
        """
        cache_backend = self.get_backend()
        try:
            item = cache_backend.get_cache_item(
                app_name=self.kwargs.get('app_name'),
                name=self.kwargs.get('item_name'),
                project=self.get_project(),
            )
        except Exception as ex:
            raise ParseError(ex)
        if not item:
            raise NotFound()
        return item


@extend_schema(
    responses={
        '200': inline_serializer(
            'UpdateTimeResponse',
            fields={'update_time': serializers.IntegerField()},
        )
    }
)
class CacheItemDateRetrieveAPIView(
    SodarcacheAPIViewMixin, SODARAPIGenericProjectMixin, APIView
):
    """
    Retrieve timestamp of the last update to a cache item. Returns 404 if cache
    item is not set.

    **URL:** ``/cache/api/retrieve/date/{Project.sodar_uuid}/{app_name}/{item_name}``

    **Methods:** ``GET``

    **Returns:**

    - ``update_time``: Update timestamp in seconds since epoch (integer)
    """

    permission_required = 'sodarcache.get_cache_value'
    project_type = PROJECT_TYPE_PROJECT
    http_method_names = ['get']

    def get(self, request, *args, **kwargs):
        cache_backend = self.get_backend()
        try:
            update_time = cache_backend.get_update_time(
                app_name=self.kwargs.get('app_name'),
                name=self.kwargs.get('item_name'),
                project=self.get_project(),
            )
        except Exception as ex:
            raise ParseError(ex)
        if not update_time:
            raise NotFound()
        return Response({'update_time': update_time}, status=200)


class CacheItemSetAPIView(
    SodarcacheAPIViewMixin, SODARAPIGenericProjectMixin, APIView
):
    """
    Create or update a cache item. Replaces an existing item with the same
    project, app name and item name if previously set. Returns 200 on both a
    successful creation and update.

    **URL:** ``/cache/api/set/{Project.sodar_uuid}/{app_name}/{item_name}``

    **Methods:** ``POST``

    **Parameters:**

    - ``data``: Full item data to be set (JSON)
    """

    http_method_names = ['post']
    permission_required = 'sodarcache.set_cache_value'
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = JSONCacheItemSerializer

    def post(self, request, *args, **kwargs):
        cache_backend = self.get_backend()
        try:
            cache_backend.set_cache_item(
                app_name=self.kwargs.get('app_name'),
                name=self.kwargs.get('item_name'),
                user=self.request.user,
                data=request.data['data'],
                data_type='json',
                project=self.get_project(),
            )
        except Exception as ex:
            raise APIException(ex)
        return Response({'detail': 'ok'}, status=200)
